// --- nServer.c v1----
#include <os.h>
#include <lauxlib.h>

typedef unsigned int       u32;

// see the links below
//http://www.omnimaga.org/index.php?topic=2255.msg58307#msg58307
//http://www.omnimaga.org/index.php?topic=2255.msg58391#msg58391
//Thanks Goplat, finally found something useful with this important call
struct event {   
	unsigned int timestamp;
	unsigned short type;
	unsigned short ascii;
	unsigned int key;
	unsigned int unknown[3];
	unsigned int modifier;
};

//static const unsigned luaL_checknumber_addrs[] = {0x106c76c8, 0x0, 0x106C6318, 0x0};  //OS 3.1
//#define _luaL_checknumber SYSCALL_CUSTOM(luaL_checknumber_addrs, int,  lua_State *, int)

static const unsigned get_event_addrs[] = {0x1021E544, 0x1021EC70, 0x1021dfb4, 0x1021e744};  //OS 3.1
#define get_event SYSCALL_CUSTOM(get_event_addrs, int,  struct event *)

static const unsigned sscanf_addrs[] = {0x10378828, 0x10378F58,0x103760E8,0x10376878};  // OS 3.1 
#define sscanf SYSCALL_CUSTOM(sscanf_addrs ,int, char *str, char *fmtin, ... )

static const unsigned disp_str_addrs[] = {0x100b30c4, 0x100b3384, 0x100b275C , 0x100b2a44 };  // OS 3.1 - 6x8 font
#define disp_str SYSCALL_CUSTOM(disp_str_addrs ,int,  const char *, int*,int )

static const unsigned ascii6x8table[] = {0x10840A4C, 0x10880AA4, 0x10821918 , 0x1087dc18 };// OS 3.1


// ---- screen variables ---
static char strbuf[200]; 
static char base64buf[200];
//static char scrbuf[30][53]; // screen character buffer 30 rows x 53 columns all calcs 
static int curX, curY; // global current cursor position
static char dashline[] =  "-----------------------------------------------------";
static char blankline[] = "                                                     ";
static char whiteline[] = "\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c";


// ---- program variables ---

static char *packetptr;
static int packno;   // packet number
static int recording;  // packet recording flag
static unsigned modifier;
static unsigned etype;


// ---- program functions ---

void ShowCursor(void) {
 disp_str("\x1e",&curX, curY);  
 --curX;
}

void fixfonttable(void){
  unsigned charindex;
  unsigned table;
   switch (*(unsigned*)(0x10000020)) {
	case 0x102F0FA0:  // 3.1.0 non-CAS
           table = ascii6x8table[0];   
           break;
	case 0x102F16D0:  // 3.1.0 CAS
           table = ascii6x8table[1];   
	   break;
	case 0x102F0A10:  // 3.1.0 non-CAS CX
           table = ascii6x8table[2];   
	   break;
	case 0x102F11A0:  // 3.1.0 CAS CX
           table = ascii6x8table[3];   
	   break;
        default:
           return; 
       }    
// fix missing vertical bar or pipe
  charindex = table + 32*(int)'|'; 
  *(int*)charindex = 4; charindex+=4;
  *(int*)charindex = 4; charindex+=4;
  *(int*)charindex = 4; charindex+=4;
  *(int*)charindex = 0; charindex+=4;
  *(int*)charindex = 4; charindex+=4;
  *(int*)charindex = 4; charindex+=4;
  *(int*)charindex = 4; 
// fix missing right brace
  charindex = table + 32*(int)'}';
  *(int*)charindex = 2; charindex+=4;
  *(int*)charindex = 1; charindex+=4;
  *(int*)charindex = 1; charindex+=4;
  *(int*)charindex = 2; charindex+=4;
  *(int*)charindex = 1; charindex+=4;
  *(int*)charindex = 1; charindex+=4;
  *(int*)charindex = 2; 
// fix missing left brace
  charindex = table + 32*(int)'{';
  *(int*)charindex = 0x8; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x8; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x8; 
// fix missing tilde
  charindex = table + 32*(int)'~';
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x8; charindex+=4;
  *(int*)charindex = 0x15; charindex+=4;
  *(int*)charindex = 0x2; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4; 
// add house symbol to DEL 
  charindex = table + 32*0x7f;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x4; charindex+=4;
  *(int*)charindex = 0xa; charindex+=4;
  *(int*)charindex = 0x11; charindex+=4;
  *(int*)charindex = 0x11; charindex+=4;
  *(int*)charindex = 0x1F; charindex+=4;   
// fix the small l
  charindex = table + 32*(int)'l';
  *(int*)charindex = 0x0C; charindex+=4;
  *(int*)charindex = 0x4; charindex+=4;
  *(int*)charindex = 0x4; charindex+=4;
  *(int*)charindex = 0x4; charindex+=4;
  *(int*)charindex = 0x4; charindex+=4;
  *(int*)charindex = 0x4; charindex+=4;
  *(int*)charindex = 0x4;
// fix the small r
  charindex = table + 32*(int)'r';
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0xb; charindex+=4;
  *(int*)charindex = 0xd; charindex+=4;
  *(int*)charindex = 0x9; charindex+=4;
  *(int*)charindex = 0x8; charindex+=4;
  *(int*)charindex = 0x8;  
// add a hatched block cursor character for 0x1f
  charindex = table + 32*0x1f;
  *(int*)charindex = 0x15; charindex+=4;
  *(int*)charindex = 0xa; charindex+=4;
  *(int*)charindex = 0x15; charindex+=4;
  *(int*)charindex = 0xa; charindex+=4;
  *(int*)charindex = 0x15; charindex+=4;
  *(int*)charindex = 0xa; charindex+=4;
  *(int*)charindex = 0x15;  
// add a double underscore cursor character for 0x1E
  charindex = table + 32*0x1E;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x0; charindex+=4;
  *(int*)charindex = 0x3F; charindex+=4;
  *(int*)charindex = 0x3F; 
// add a vertical bar cursor character for 0x1D
  charindex = table + 32*0x1D;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; charindex+=4;
  *(int*)charindex = 0x10; 
// white block character for 0x1c
  charindex = table + 32*0x1c;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; charindex+=4;
  *(int*)charindex = 0x3f; 
}

void scrollregion(int rows, int cursorX, int cursorY){  // number of rows to scroll - usually 1
       int x= 0, y=2;  // keep the packet lines at top
       if(is_cx){
        memmove(SCREEN_BASE_ADDRESS+2*5120, SCREEN_BASE_ADDRESS+2*5120 + 5120*rows, 153600 - 5*5120- 5120*rows);
        for(y=26-rows+1;y<=26;++y){   // clear the scrolled in region
         x = 0; 
         disp_str(blankline,&x,y);  
        }  
       }else{ // classic NSpire
        memmove(SCREEN_BASE_ADDRESS+2*1280, SCREEN_BASE_ADDRESS+2*1280 + 1280*rows, 38400 - 5*1280 - 1280*rows);
        for(y=26-rows+1;y<=26;++y){   // clear the scrolled in region  
         x = 0; 
         disp_str(blankline,&x,y); 
        }
       } 
       curX = cursorX; curY = cursorY;  // move cursor when done
}

void  startPacket(void){ 
 int x,y;
  packetptr = (char*)(SCREEN_BASE_ADDRESS);  // reset pointer
// x = 0; y = 0;
// disp_str(whiteline,&x,y);  // clear old packet
// x = 0; y = 1;
// disp_str(whiteline,&x,y);  // clear old packet 2nd line
 *packetptr = '\0';  // for strcat
 recording = 1;  // flag on
}

void stopPacket(void){
 char num[20];
 int len = strlen(packetptr);
 if(len % 2 == 1){  // Jim Bauwens packet scheme in C
   *(packetptr + len + 1)  = '\xff';  // fill white "stop" pixel
   *(packetptr + len + 2)  = '\xff';  // fill white "stop" pixel
 }else{                        // even length string
   *(packetptr + len + 0)  = '\xff';  // fill white "stop" pixel
   *(packetptr + len + 1)  = '\xff';  // fill white "stop" pixel      
 }
// sprintf(num,"%d",packno);
//? strcpy(packetptr+strlen(packetptr)+1,num);  // append packet number after string null
 recording = 0; // flag off
 ++packno;
}

void  startluaPacket(void){ 
 int x,y;
  packetptr = (char*)(SCREEN_BASE_ADDRESS);  // reset pointer
  x = 0; y = 0;
// disp_str("\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c\x1c",&x,y);  // clear old packet
 *packetptr = '\0';  // for strcat
}

void stopluaPacket(int size){
 int x,y;
 char num[20];
 sprintf(num,"%d",packno);
 strcpy(packetptr+size,num);  // append packet number after string null
 x = 15; y = 2;
 disp_str(num,&x,y);  // display packet number on menu bar 
//      lua_pushstring(L,num);
//      lua_setglobal(L,"nsrvrmsg");
 ++packno;
}

void DrawStr(char *str){
 if(recording){
  strcat(packetptr,str);
 }
 if (curY > 26){
  scrollregion(1,0,26);  // scroll one line up if cursor at bottom of region
 }
 if (*(char*)(str + strlen(str)-1) == '\n'){ 
  *(char*)(str + strlen(str)-1) = '\0';
  disp_str(str,&curX,curY);
  curX=0;++curY;
  if (curY > 26) scrollregion(1,0,26);  // scroll one line up if cursor at bottom of region  
 } else {
  disp_str(str,&curX,curY);
 }
}

int getch(void) {
	struct event event;
	do {
           idle();
	   get_event(&event);
	} while (!(event.type == 0x208 || event.type == 0x8) ); 
        modifier = event.modifier;
        etype = event.type;
	return event.ascii;
}

char *gets(char *buf) {
	char ch,chs[4],g[20];
        int len,x,y,sinkeybytes,ackbytes;
        buf[0] = '\0';
        chs[0] = '\0';
        chs[1] = '\0';
        chs[2] = '\0';
        chs[3] = '\0';
	do {
	  ch = (char)getch();
          if (etype == 8 && ch == '\b' && strlen(buf) > 0 && curX > 0){  // backspace - rubout
           disp_str("\x01",&curX,curY); // rubout the cursor
           --curX; --curX;
           disp_str("\x1E",&curX,curY); // overwrite the character with the cursor
           --curX;
           len = strlen(buf);
           buf[--len] = '\0';
           continue;
          }
          if ((int)ch >= 0x20 && (int)ch < 0x80){  // printable ascii char
//           ch = toupper(ch);  // uppercase for this font looks better
           chs[0] = ch;
           x = 0; y= 28;
           sprintf(g,"Mod:%x typ:%x",modifier,etype); disp_str(g,&x,y);
           if (curY > 26) scrollregion(1,0,26);
           disp_str(chs,&curX,curY);   // echo the key
           ShowCursor();  
           strcat(buf,chs);
           continue;
          }
//   ---- special char sequences from nRemote -----
          if (etype == 0x208 && (ch == '\0' || ch == 't')){  // 0,0x74,0: message delimiter
           sinkeybytes <<=8 ; sinkeybytes |= (int)ch;
           if(sinkeybytes == 0x7400){
            sinkeybytes = 0;
            x = 0; y= 28;
            strcpy(g,"MSG        "); disp_str(g,&x,y);            
           }
          }
          if (etype == 0x208 && (ch == '\0' || (int)ch == 124 || (int)ch == 149)){  // 124,149,0 ACKNOWLEDGE SEQUENCE
           ackbytes <<=8 ; ackbytes |= (int)ch;
           if(sinkeybytes == 0x7C9500){
            ackbytes = 0;
            x = 0; y= 28;
            strcpy(g,"ACK        "); disp_str(g,&x,y);            
           }
          }
	} while (ch != '\n' && ch != '\r' && ch != (char)27);   // return key(10) or Enter key(13) or ESC key(27)
        disp_str(" ",&curX,curY); // erase cursor
        curX = 0; ++curY;    // advance to next line
        if (curY > 26){
         scrollregion(1,0,26);  // scroll one line up if cursor at bottom of region
        } 
        ShowCursor();       
        if ((int)ch ==27){strcat(buf,"\x1b");} 
	return buf;
}

void hexdump(unsigned address){    // modded from nspire_emu: debug.c
	unsigned start = address;   
	unsigned end = address + 0x7F;
        unsigned char *ptr;
	unsigned row, col;
        char buf[80];

	for (row = start & ~0x7; row <= end; row += 0x8) {
                ptr = (unsigned char*) row ;           
                sprintf(buf,"%08X  ", row);
		DrawStr(buf);
		for (col = 0; col < 0x8; col++) {
			address = row + col;
			if (address < start || address > end)
				DrawStr("  ");
			else {
                                sprintf(buf,"%02X", ptr[col]);
				DrawStr(buf);
                        }
			DrawStr(" ");
		}
		DrawStr("  ");
		for (col = 0; col < 0x8; col++) {
			address = row + col;
			if (address < start || address > end)
				DrawStr(" ");
			else if ((int)ptr[col] < 0x20 || (int)ptr[col] > 0x7f)
				DrawStr(".");
			else{
                                sprintf(buf,"%s", (char*)(ptr+col)); buf[1] = '\0';
				DrawStr(buf);
                        }
		}
                buf[0] = '\n' ; buf[1] = '\0';
		DrawStr(buf);
                
	}
}


void clearscreen(void){
        int x= 0, y=0;       
        for(y=0;y<30;++y){ 
         x = 0;
         if(y!=0 && y!=1 && y!=27){  
          disp_str(blankline,&x,y);
         } 
         if(y==0 || y==1){
          disp_str(whiteline,&x,y);
         }
         if(y==27){
          disp_str(dashline,&x,y);
         }
        }  
        curX = 0; curY = 2; 
        ShowCursor(); 
}

double dswap(double v)  // convert a Yagarto double to a TI double, might not need in later version of Ndless
{
    union { 
        unsigned long long i;
        double  d;
    } conv;
    conv.d = v;
    conv.i = (conv.i << 32) | (conv.i >> 32);
    //printf("%08x %08x %lf\n",(unsigned)(conv.i >> 32), (unsigned)conv.i,  conv.d);
    return conv.d;
}

static int lua_embedprint (lua_State *L){  
    int i,len,endian;
    char ch;
    const char *param = luaL_checkstring(L, 1);
    endian = luaL_checkint(L, 2);
    startluaPacket();
    strcat(strbuf,param);
    if (endian == 1){
      for(i=0;i<200;i+=2){ch = strbuf[i]; strbuf[i]  = strbuf[i+1]; strbuf[i+1] = ch;}
     memcpy(packetptr,strbuf,strlen(param));
    } else {
     strcat(packetptr,param); 
    } 
    sprintf(strbuf,"%d",packno);
    len = strlen(param);
    if(len % 2 == 1){  // Jim Bauwens packet scheme in C
     *(packetptr + len + 1)  = '\xff';  // fill white "stop" pixel
     *(packetptr + len + 2)  = '\xff';  // fill white "stop" pixel
     len = len + 2;
    }else{                        // even length string
     *(packetptr + len + 0)  = '\xff';  // fill white "stop" pixel
     *(packetptr + len + 1)  = '\xff';  // fill white "stop" pixel  
     len = len + 1;   
    }
    stopluaPacket(len); 
    return 0;
}

static int lua_peek(lua_State* L) {
  unsigned addr;
  int type = lua_type(L, 1);
  if (type != LUA_TSTRING){
   lua_pushstring(L,"nServer.peek, address not a stringed number");
   lua_error(L);
  }  
  addr =  strtoul(luaL_checkstring(L, 1),NULL,0);
  int size = luaL_checkint(L, 2);
  lua_pushlstring(L, (void *)addr, size);
  return 1;
}

static int lua_poke(lua_State* L) {
  unsigned addr;
  size_t size ; 
  int type = lua_type(L, 1);
  if (type != LUA_TSTRING){
   lua_pushstring(L,"nServer.poke, address not a stringed number");
   lua_error(L);
  }  
  addr =  strtoul(luaL_checkstring(L, 1),NULL,0);
  const char* str = luaL_checklstring(L, 2, &size);
  memcpy((void *)addr, (void *)str, size);
  return 0;
}

static int lua_strptr(lua_State* L) { // get a pointer to a Lua string, stop garbage collection when using
  int type = lua_type(L, 1);
  char str[12];
  if (type != LUA_TSTRING){
   lua_pushstring(L,"nServer.Ptr, not a string");
   lua_error(L);
  }
  const char *s = lua_tostring(L,1);  
  sprintf(str,"0X%08X",(unsigned)s);
  lua_pushstring(L,str);
  lua_pushinteger(L, (int)s);
  return 2;  // return as a string and Lua number 
}

static int lua_malloc(lua_State* L) {
  size_t size ; 
  size = luaL_checkint(L, 1);
  lua_pushinteger(L, (int)malloc(size));
  return 1;
}

static int lua_free(lua_State* L) {
  int addr; 
  addr = luaL_checkint(L, 1);
  free((void*)addr);
  return 0;
}

static int int2num(lua_State* L) { // convert a C int to lua number 
  double dd;
  int type = lua_type(L, 1);
  if (type != LUA_TSTRING){
   lua_pushstring(L,"nServer.int2num, not a string");
   lua_error(L);
  }
  dd = (double) *(unsigned*)lua_tostring(L,1);
  //if ( dd > 2147483647.0d ) {dd -= 4294967296.0d;}  // makes signed
  lua_pushfstring(L,"%f",dd);   // return as unsigned
  return 1;
}
// doubles can represent both signed and unsigned ints
static int num2int(lua_State* L) {  // convert a Lua number to a C integer
  const char *dstr = luaL_checkstring(L,1);  // luaL_checkstring converts doubles if nececcary
  double dd;  // lua_tonumber[106C6318,old106aa80c] luaL_checknumber[106C76C0,old106c78b4] lua_pushnumber[106c78b4,106c5340]
  sscanf(dstr,"%lf",&dd);
  if ( dd < 0.0d ) {dd += 4294967296.0d;}  // makes unsigned representation
  int n = (int)(unsigned)dd;
  lua_pushlstring(L,(char*)&n, 4);   
  return 1;
}

static int str2ul(lua_State* L) { // convert a stringed lua unsigned number to a C unsigned integer
  int type = lua_type(L, 1);
  if (type != LUA_TSTRING){
   lua_pushstring(L,"nServer.str2ul, not a string");
   lua_error(L);
  }
  type = strtoul(lua_tostring(L,1),NULL,0);
  lua_pushlstring(L,(char*)&type,4);   
  return 1;
}

static int str2l(lua_State* L) { // convert a stringed lua signed number to a C  integer
  int type = lua_type(L, 1);
  if (type != LUA_TSTRING){
   lua_pushstring(L,"nServer.str2l, not a string");
   lua_error(L);
  }
  type = strtol(lua_tostring(L,1),NULL,0);
  lua_pushlstring(L,(char*)&type,4);   
  return 1;
}

static int strto32bit(const char *s) {
 int n;
 if(strstr(s,"-") || strstr(s,"+")){
  n = strtol(s,NULL,0);   // signed integers -2147483648<=x<=2147483647[0x80000000 <= x <= 7fffffff]
 }else{
  n = strtoul(s,NULL,0);  // unsigned integers 0<=x<=0xffffffff
 }
 return n;
}

static int lua_externcall(lua_State* L) {
  unsigned a[10],ret,type;
  int i,top;
  top = lua_gettop(L);
  unsigned funcptr = (unsigned)strto32bit(lua_tostring(L, 1));
  for(i=2;i<=top;++i){
   type = lua_type(L, i);
   if (type == LUA_TSTRING){
    a[i-2] = (unsigned)strto32bit(lua_tostring(L, i));
   } 
   if (type == LUA_TNUMBER){
    a[i-2] = (unsigned)strto32bit(luaL_checkstring(L, i));
   }   
  }
  ret = ((unsigned(*)(u32, u32, u32, u32, u32, u32))funcptr)(a[0],a[1],a[2],a[3],a[4],a[5]);
  lua_pushinteger(L, (int)ret);
  return 1;
}

static int bitops(lua_State* L) { // bitops on lua hex strings, results to a C integer & hex string
  char hex[16];
  const char *aa = luaL_checkstring(L,1);
  const char *bb = luaL_checkstring(L,2);
  const char *ch = luaL_checkstring(L,3);
  int a = (int)strtoul(aa,NULL,0);
  int b = (int)strtoul(bb,NULL,0);
  switch((int)ch[0]){ 
   case (int)'|':
    lua_pushinteger(L, a|b);
    sprintf(hex,"0X%X",a|b); 
    break;
   case (int)'^':
    lua_pushinteger(L, a^b);
    sprintf(hex,"0X%X",a^b); 
    break;   
   case (int)'&':
    lua_pushinteger(L, a&b);
    sprintf(hex,"0X%X",a&b);
    break;
   case (int)'~':
    lua_pushinteger(L, ~a);
    sprintf(hex,"0X%X",~a);
    break;
   case (int)'<':
    lua_pushinteger(L, a<<b);
    sprintf(hex,"0X%X",a<<b);
    break;
   case (int)'>':
    lua_pushinteger(L, a>>b);
    sprintf(hex,"0X%X",a>>b);
    break;
   default:
    lua_pushstring(L,"bitops, 3rd arg undefined");
    lua_error(L);
  }  
  lua_pushfstring(L,hex); 
  return 2;
}

void printhelp(void){
char helpstr[][80] = { 
  "<ESC> | exit - quits nServer\n",
  "? | help -  shows this help\n", 
  "send -  send a message to nRemote\n",
  "rw <address>- read a dword at an address\n",
  "ww <address> <dword>- write a dword at an address\n",
  "cls -  clearscreen\n",
  "dump <address>- hex/ascii dump\n" };

  startPacket(); 
  DrawStr(helpstr[0]); 
  DrawStr(helpstr[1]);
  DrawStr(helpstr[2]);  
  DrawStr(helpstr[3]);    
  DrawStr(helpstr[4]); 
  DrawStr(helpstr[5]); 
  DrawStr(helpstr[6]); 
  stopPacket(); 
}

void sendmessage(void){
 startPacket();
 strcpy(packetptr,strbuf);
 stopPacket();
}

void ww(unsigned addr, unsigned dword){  // write word, word==32bits
 char cmd[] = "addresses must be 4byte aligned\n";
 if(addr % 4 != 0){
  startPacket(); 
  DrawStr(cmd);
  stopPacket();
 }else{
  *(unsigned*)addr = dword;
 }
}

void rw(unsigned addr){    // read word, word==32bits
 unsigned dword;
 char cmd[] = "addresses must be 4byte aligned\n";
 if(addr % 4 != 0){
  startPacket(); 
  DrawStr(cmd);
  stopPacket();
 }else{
  dword = *(unsigned*)addr;
  sprintf(strbuf,"0x%08X\n",dword);
  startPacket(); 
  DrawStr(strbuf);
  stopPacket();  
 }
}

int parseinput(void){   
 int arg1, arg2, arg3,arg4, scanned;
 char cmd[80];
 gets(strbuf);

 if(strstr(strbuf,"exit")  || strstr(strbuf,"\x1b") ){
  return 0;
 } 
 if(strstr(strbuf,"help")  || strbuf[0] == '?'){
  printhelp();
  ShowCursor();  
  return 1;
 } 
 if(strstr(strbuf,"send") == strbuf  ) {
  sendmessage();
  ShowCursor(); 
  return 2;
 }
 if(strstr(strbuf,"rw") ==  strbuf ) {
  scanned = sscanf(strbuf,"rw %x",&arg1);
  if(scanned == 1){
   rw(arg1);
   ShowCursor();
  } 
  return 3;
 }
 if(strstr(strbuf,"ww") ==  strbuf) {
  scanned = sscanf(strbuf,"ww %x %x",&arg1,&arg2);
  if(scanned == 2){
   ww(arg1,arg2);
   ShowCursor();
  } 
  return 4;
 }
 if(strstr(strbuf,"cls") ) {
  clearscreen();
  return 5;
 }
 if(strstr(strbuf,"dump")  ) {
  scanned = sscanf(strbuf,"dump %x",&arg1);
  startPacket();
  if(scanned == 1){
   hexdump((unsigned)arg1);
   ShowCursor();
  }
  stopPacket();
  return 6;
 }
 // not in the list
  sprintf(cmd,"Command not found: %s\n",strbuf);
  startPacket();
  DrawStr(cmd);  
  ShowCursor();
  stopPacket(); 
  return -1; 
}

static const luaL_reg lualib[] = {
	{"MsgToScreen", lua_embedprint},
	{"peek", lua_peek},
	{"poke", lua_poke},
	{"call", lua_externcall},
	{"malloc", lua_malloc},
	{"free", lua_free},
	{"Ptr",  lua_strptr},
	{"int2num",  int2num},
	{"num2int",  num2int},
	{"str2l",  str2l},
	{"str2ul",  str2ul},
	{"bitops",  bitops},
	{NULL, NULL}
};

int main(void) {
        int cmdstat = 1;
        packno = 0; // init packet counter
        recording = 0; // init packet recording flag  
        fixfonttable();
	lua_State *L = nl_lua_getstate(); 
	if (!L) { // ----- called as an interactive program
         clearscreen(); 
// ----- enter interactive loop ------       
	 while(cmdstat) {
           cmdstat = parseinput();
	 }
        } else { // ----- called as Lua module
	 nl_relocdata((unsigned*)lualib, (sizeof(lualib) / sizeof(unsigned*)) - 2);
	 luaL_register(L, "nServer", lualib);      
        } 
       return 0;
}